@@ -133,7 +133,7 @@ utf8_wordcomplete_open (GokWordComplete *complete, const gchar *directory)
 
 	g_free (contents);
 
-	if (dictionary_lines == NULL) 
+	if (dictionary_lines == NULL || dictionary_lines[0] == NULL) 
 	{
 		return FALSE;
 	}
