@@ -88,6 +88,10 @@
 
 static gboolean respawn_on_segv = TRUE;
 
+static gboolean queue_input_events = TRUE;
+
+static guint input_idle_handler = 0;
+
 static GokApplication *_gok_app = NULL;
 
 /* pointer to the first keyboard in the list of keyboards */
@@ -129,6 +133,7 @@ typedef struct _GokArgs {
 	char *geometry;
 	int geometry_bitmask;
 	gboolean is_login;
+	gboolean debug_nameless;
 	int remember_geometry;
 	int display_settings_dialog;
 	gboolean use_extras;
@@ -141,6 +146,11 @@ typedef struct _GokArgs {
 
 static GokArgs gok_args;
 
+static struct {
+    gint n_axes;
+    int *axes;
+} gok_motion_event = {0, 0x0000};
+
 /* Private functions */
 static void gok_main_read_rc (void);
 static void gok_main_initialize_access_methods (GokArgs *args);
@@ -241,6 +251,15 @@ static struct poptOption options[] = {
 		NULL
 	},
 
+	{
+		"debugnameless", 
+		'\0', 
+		POPT_ARG_NONE, 
+		&gok_args.debug_nameless, 
+		0,
+		N_("GOK will display keys for nameless GUI. Useful for debugging."),
+		NULL
+	},
 
 	{
 		"remembergeometry", 
@@ -338,6 +357,7 @@ gok_args_init (GokArgs *args)
 	args->geometry = NULL;
 	args->geometry_bitmask;
 	args->is_login = 0;
+	args->debug_nameless = FALSE;
 	args->remember_geometry = FALSE;
 	args->display_settings_dialog = FALSE;
 	args->use_extras = FALSE;
@@ -785,14 +805,10 @@ gok_main_open(gint argc, gchar *argv[])
 	   we use the login keyboard, otherwise the regular main keyboard */
 	if (gok_args.mainkeyboardname == NULL) {
 		if (gok_args.is_login) {
-			gok_args.mainkeyboardname = (gchar*) 
-				g_malloc (strlen (GOKLOGINKEYBOARDNAME) + 1);
-			strcpy (gok_args.mainkeyboardname, (gchar *) GOKLOGINKEYBOARDNAME);
+			gok_args.mainkeyboardname = (gchar*)  g_strdup ((gchar *) GOKLOGINKEYBOARDNAME);
 		}
 		else {
-			gok_args.mainkeyboardname = (gchar*) 
-				g_malloc (strlen (GOKMAINKEYBOARDNAME) + 1);
-			strcpy (gok_args.mainkeyboardname, (gchar *) GOKMAINKEYBOARDNAME);
+			gok_args.mainkeyboardname = (gchar*) g_strdup((gchar *) GOKMAINKEYBOARDNAME);
 		}
 	}
 
@@ -1096,6 +1112,40 @@ gok_main_object_state_listener (Accessible* pAccessible)
 	}
 }	
 
+/** 
+ * gok_main_check_input_queue (gpointer data):
+ *
+ * Check the extended input event queue and act on the most recent motion event,
+ * and discard other motion events.
+ * 
+ * returns: FALSE as specified in g_idle_add ().
+ **/
+static gboolean gok_main_idle_input (gpointer data)
+{
+    input_idle_handler = 0;
+    if ((gok_motion_event.n_axes > 0) && (gok_motion_event.n_axes < 4)) {
+	gok_scanner_input_motion (gok_motion_event.axes, gok_motion_event.n_axes);
+    }
+    return FALSE;
+}
+
+void
+gok_main_queue_input_event (int *motion_data, gint n_axes)
+{
+    if (n_axes > 0 && n_axes < 4) {
+	if (gok_motion_event.axes)
+	    g_free (gok_motion_event.axes);
+	gok_motion_event.axes = g_new (int, n_axes);
+	gok_motion_event.n_axes = n_axes;
+	memcpy (gok_motion_event.axes, motion_data, n_axes * sizeof (int));
+	if (input_idle_handler == 0) {
+	    input_idle_handler = g_idle_add (gok_main_idle_input, NULL);
+	}
+    }
+    else
+	g_warning ("Motion Event has an invalid number of axes!");
+}
+
 
 /**
 * gok_main_motion_listener
@@ -1105,6 +1155,9 @@ gok_main_object_state_listener (Accessible* pAccessible)
 * @timestamp: long int, ignored
 *
 * This handler is called each time there is a motion event from the connected input device.
+* We queue the input event for processing during idle, as the rate of input events
+* may exceed our ability to process them, otherwise, and we only need to consider the most
+* recent motion event.
 */
 void gok_main_motion_listener (gint n_axes, int *motion_data, long mods, long timestamp)
 {
@@ -1118,7 +1171,10 @@ void gok_main_motion_listener (gint n_axes, int *motion_data, long mods, long ti
 			mods, timestamp);
 		gok_log_leave ();
 #endif
-		gok_scanner_input_motion (motion_data, n_axes);
+		if (queue_input_events)
+		    gok_main_queue_input_event (motion_data, n_axes);
+		else
+		    gok_scanner_input_motion (motion_data, n_axes); 
 
 		if (gok_data_get_drive_corepointer ()) {
 			Display *display;
@@ -1580,6 +1636,9 @@ gboolean gok_main_ds (GokKeyboard* pKeyboard)
 	/* stop any key flashing */
 	gok_feedback_timer_stop_key_flash();
 	
+	/* we have branched, so we should make sure 'repeat' is off */
+	gok_repeat_disarm ();
+
 	/* stop the scanning process */
 	gok_scanner_stop();
 
@@ -1917,7 +1976,7 @@ gok_main_read_keyboards_from_dir (const char *directory, GokKeyboard *keyboard)
 	gchar* filename;
 	gchar* complete_path;
 	GList* langlist = NULL;
-	gchar dirtemp[1024]; /* no reliable way to avoid magic # here */
+	gchar *dirtemp;
 
 	gok_log_enter();
 
@@ -1928,22 +1987,20 @@ gok_main_read_keyboards_from_dir (const char *directory, GokKeyboard *keyboard)
 	if (langlist != NULL) {
 		gint i = 0;
 		guint l = g_list_length(langlist);
-		while ((pDirectoryKeyboards == NULL) && (i < l)) {
-			strncpy(dirtemp, directory, 950);
-			strncat(dirtemp, G_DIR_SEPARATOR_S, 1);
-			strncat(dirtemp, g_list_nth_data(langlist, i),70);
+		while ((i < l)) {
+			dirtemp = g_build_filename (directory, g_list_nth_data(langlist, i), NULL);
 			pDirectoryKeyboards = opendir (dirtemp);
+			if (pDirectoryKeyboards) 
+				break;
+			g_free (dirtemp);
 			i++;
 		}
 	}
 
 	if (pDirectoryKeyboards == NULL) { /* no locale specific directory found */
 		pDirectoryKeyboards = opendir (directory); /* use default */
+		dirtemp = g_strdup (directory);
 	}
-	else {
-		directory = dirtemp; /* used later */
-	}
-		
 
 	/* get a list of all the files from the given directory name */
 	if (pDirectoryKeyboards == NULL)
@@ -1968,7 +2025,7 @@ gok_main_read_keyboards_from_dir (const char *directory, GokKeyboard *keyboard)
 		     && (strcmp(filename + (strlen(filename)-4), ".kbd") == 0))
 		{
 			/* read the keyboard file and add to list */
-                        complete_path = g_build_filename (directory, filename,
+                        complete_path = g_build_filename (dirtemp, filename,
                                                           NULL);
 			gok_log ("complete_path = %s", complete_path);
 			keyboard = gok_append_keyboard_from_file (keyboard, complete_path);
@@ -1979,6 +2036,7 @@ gok_main_read_keyboards_from_dir (const char *directory, GokKeyboard *keyboard)
 		}
 	}
  	g_free (pDirectoryKeyboards);
+	g_free (dirtemp);
 	gok_log_leave();
 	return pKeyboardFirst;
 }
@@ -2496,6 +2554,9 @@ gok_main_close()
 	GokKeyboard* pKeyboard;
 	GokKeyboard* pKeyboardTemp;
 
+	/* get rid of input idle handler */
+	if (input_idle_handler) g_source_remove (input_idle_handler);
+
 	/* restore the corepointer, if it was changed... */
 	gok_input_restore_corepointer ();
 
@@ -2561,10 +2622,9 @@ gok_main_close()
 void gok_main_display_error (gchar* ErrorString)
 {
 	GtkWidget* pDialog;
-	gchar buffer [1024];
+	gchar* buffer;
 	
-	strcpy (buffer, _("Sorry, GOK can't run because:\n"));
-	strcat (buffer, ErrorString);
+	buffer = g_strdup_printf (_("Sorry, GOK can't run because:\n%s"), ErrorString);
 	
 	pDialog = gtk_message_dialog_new (NULL,
 		GTK_DIALOG_DESTROY_WITH_PARENT,
@@ -2575,6 +2635,7 @@ void gok_main_display_error (gchar* ErrorString)
 	gtk_window_set_title (GTK_WINDOW (pDialog), _("GOK Fatal Error"));
 	gtk_dialog_run (GTK_DIALOG (pDialog));
 	gtk_widget_destroy (pDialog);
+	g_free (buffer);
 }
 
 /**
@@ -2862,7 +2923,9 @@ void check_accessibility_cb ( GObject* o, gpointer* data )
 {
 	GnomeClient *client;
 
-	gtk_widget_destroy(acd);	
+	if (acd) { 
+	    gtk_widget_destroy(acd); acd = NULL; 
+	}	
 	
 	if (strcmp((gchar*)data,"logout") == 0) {
 		gok_gconf_set_bool ( gok_data_get_gconf_client(),
@@ -2891,14 +2954,18 @@ gok_main_help_cb (gpointer data)
 	GError *error = NULL;
 	gchar *helpuri = data;
 	/* TODO: detect error launching help, and give informative message */
-	gnome_help_display_desktop (NULL, "gok", helpuri, NULL, &error);
+	/* TODO: remove hack for link below, if and when this method is called from more than one dialog */
+
+	gnome_help_display_desktop (NULL, "gnome-access-guide", helpuri, "sysadmin-37", &error);
 }
 
 void
 gok_main_close_warning ()
 {
-        gtk_widget_destroy (_corepointer_warning);
-	_corepointer_warning = NULL;
+        if (_corepointer_warning) {
+                gtk_widget_destroy (_corepointer_warning);
+	        _corepointer_warning = NULL;
+        }
 } 
 
 void
@@ -2944,6 +3011,7 @@ gok_main_warn_corepointer (gboolean always, gboolean configuring, gboolean opt_o
 gboolean
 gok_main_warn (gchar *message, gboolean always, GCallback cancelfunc, GCallback okfunc, gboolean opt_out)
 {
+
 	if (always || gok_scanner_current_state_uses_corepointer ()) 
 	{
 		GtkWidget*	button;
@@ -2965,7 +3033,7 @@ gok_main_warn (gchar *message, gboolean always, GCallback cancelfunc, GCallback
 			gtk_container_add (GTK_CONTAINER (GTK_DIALOG(_corepointer_warning)->action_area),
 					   button);
 			g_signal_connect_swapped (G_OBJECT (button), "clicked", 
-						  G_CALLBACK (gok_main_help_cb), "gok.xml");
+						  G_CALLBACK (gok_main_help_cb), "sysadmin.xml");
 			
 			button = gtk_button_new_from_stock(GTK_STOCK_PREFERENCES);
 			gtk_container_add (GTK_CONTAINER (GTK_DIALOG(_corepointer_warning)->action_area),
@@ -3063,3 +3131,9 @@ gok_main_check_accessibility ()
 		
 	}
 }
+
+gboolean
+gok_main_get_debug_nameless ()
+{
+	return gok_args.debug_nameless;
+}
